# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=unexpected-keyword-arg

"""Module for easily getting IBMQ systems"""

import warnings
from kaleidoscope.errors import KaleidoscopeError
from ._account import Account
from .filters.backend import (BackendCollection, NumQubits, HasPulse,
                              QVCompare, IsOperational,
                              MaxCircuits, MaxShots, Reservations)


class KaleidoscopeSystemService():
    """The service class for systems.

    All attributes are dynamically attached to
    this class.

    This class is much simpler than the simulators
    one because there is no processing that needs
    to be done async.

    Allows for filtering backends by the following:

    - `num_qubits` : Filter against an integer number of qubits.

    - `open_pulse` : Boolean filter for open_pulse capability.

    - `quantum_volume` : Filter against an integer number for quantum volume.

    - `operational` : Boolean filter for system operational status.

    - `max_circuits` : Filter against integer number of max circuits.

    - `max_shots` : Filter against integer number of max shots.
    """
    def __init__(self):
        self._all_added_attr = None
        self._default_added_attr = None
        self._default_added_backends = None
        self._default_provider = Account.get_default_provider()
        _system_loader(self)

    def __call__(self, name=None):
        """Return all backends that satisfy the given criteria.

        If no criteria passed then returns all systems.

        Parameters:
            name (str or list): System name(s).

        Returns:
            IBMQBackend: A single backend instance if only one.
            BackendCollection: List of specified backends if more than one.

        Raises:
            KaleidoscopeError: No matching backends.
        """
        if name is not None:
            if not isinstance(name, list):
                out = [back for back in self._default_added_backends if back.name() == name]
                if any(out):
                    return out[0]
                raise KaleidoscopeError('No matching backend name.')
            out = []
            for nm in name:
                out += [back for back in self._default_added_backends if back.name() == nm]
            if any(out):
                return BackendCollection(out)
            raise KaleidoscopeError('No matching backend names.')
        return BackendCollection(self._default_added_backends)

    def __getattr__(self, name):
        if name == 'num_qubits':
            out = NumQubits(self._default_added_backends)
        elif name == 'open_pulse':
            out = HasPulse(self._default_added_backends)
        elif name == 'quantum_volume':
            out = QVCompare(self._default_added_backends)
        elif name == 'operational':
            out = IsOperational(self._default_added_backends)
        elif name == 'max_circuits':
            out = MaxCircuits(self._default_added_backends)
        elif name == 'max_shots':
            out = MaxShots(self._default_added_backends)
        elif name == 'reservations':
            out = Reservations(self._default_added_backends)
        else:
            raise AttributeError("BackendCollection does not have attr '{}'.".format(name))
        return out

    def _refresh(self):
        """Refresh the service in place.
        """
        for key in self._default_added_attr:
            try:
                del self.__dict__[key]
            except AttributeError:
                pass
        self._all_added_attr = []
        self._default_added_attr = []
        self._default_added_backends = []
        self._default_provider = Account.get_default_provider()
        delattr(self, 'ALL')
        _system_loader(self)


class KaleidoscopeSystemDispatcher():
    """Contains all the backend instances
    corresponding to the providers for a given
    system.

    All attributes are dynamically attached to
    this class.

    Allows for filtering backends by the following:

    - `num_qubits` : Filter against an integer number of qubits.

    - `open_pulse` : Boolean filter for open_pulse capability.

    - `quantum_volume` : Filter against an integer number for quantum volume.

    - `operational` : Boolean filter for system operational status.

    - `max_circuits` : Filter against integer number of max circuits.

    - `max_shots` : Filter against integer number of max shots.
    """
    def __call__(self, name=None, hub=None, group=None, project=None):
        """Return all backends that satisfy the given criteria.

        If no criteria passed then returns all systems.

        Parameters:
            name (str or list): System name.
            hub (str or list): Specified hub.
            group (str or list): Specified group.
            project (str or list): Specified project.

        Returns:
            BackendCollection: List of specified backends.

        Raises:
            KaleidoscopeError: No matching backends.
        """
        filtered_backends = self._added_backends
        if name is not None:
            _temp = []
            if not isinstance(name, list):
                name = [name]
            for nm in name:
                _temp += [back for back in self._added_backends
                          if back.name() == nm and back in filtered_backends]
            filtered_backends = _temp

        if hub is not None:
            _temp = []
            if not isinstance(hub, list):
                hub = [hub]
            for hb in hub:
                _temp += [back for back in self._added_backends
                          if back.hub == hb and back in filtered_backends]
            filtered_backends = _temp

        if group is not None:
            _temp = []
            if not isinstance(group, list):
                group = [group]
            for gp in group:
                _temp += [back for back in self._added_backends
                          if back.group == gp and back in filtered_backends]
            filtered_backends = _temp

        if project is not None:
            _temp = []
            if not isinstance(project, list):
                project = [project]
            for pt in project:
                _temp += [back for back in self._added_backends
                          if back.project == pt and back in filtered_backends]
            filtered_backends = _temp

        if not any(filtered_backends):
            raise KaleidoscopeError('No matching systems found.')
        return BackendCollection(filtered_backends)

    def __getattr__(self, name):
        if name == 'num_qubits':
            out = NumQubits(self._added_backends)
        elif name == 'open_pulse':
            out = HasPulse(self._added_backends)
        elif name == 'quantum_volume':
            out = QVCompare(self._added_backends)
        elif name == 'operational':
            out = IsOperational(self._added_backends)
        elif name == 'max_circuits':
            out = MaxCircuits(self._added_backends)
        elif name == 'max_shots':
            out = MaxShots(self._added_backends)
        elif name == 'reservations':
            out = Reservations(self._added_backends)
        else:
            raise AttributeError("BackendCollection does not have attr '{}'.".format(name))
        return out


def _system_loader(service):
    """Attaches system dispatchers to the service
    """
    if not any(Account.providers()):
        try:
            Account.load_account()
        except Exception:  # pylint: disable=broad-except
            pass
    systems = _get_ibmq_systems()
    all_added_attr = []
    default_added_attr = []
    default_added_backends = []
    default_provider = service._default_provider.split('//') if service._default_provider else []

    setattr(service, 'ALL', KaleidoscopeSystemDispatcher())
    setattr(service.ALL, '_added_backends', [])
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for name, back_list in systems.items():
            reference = back_list[0]
            num_qubits = reference._configuration.num_qubits
            system_name = "{}𖼯{}Q𖼞".format(name, num_qubits)
            all_dispatcher = KaleidoscopeSystemDispatcher()
            for backend in back_list:
                hub = backend.hub
                group = backend.group
                project = backend.project
                pro_str = "{}_{}_{}".format(hub, group, project)
                pro_str = pro_str.replace('-', 'ー')
                setattr(all_dispatcher, pro_str, backend)
                service.ALL._added_backends.append(backend)
                # is backend in default provider
                if [hub, group, project] == default_provider:
                    setattr(service, system_name, backend)
                    default_added_attr.append(system_name)
                    default_added_backends.append(backend)

            setattr(service.ALL, 'get_'+system_name, all_dispatcher)
            all_added_attr.append(system_name)

    service._all_added_attr = all_added_attr
    service._default_added_attr = default_added_attr
    service._default_added_backends = default_added_backends


def _get_ibmq_systems():
    """Get instances for all IBMQ systems that the user has access to.

    Returns:
        dict: A dict of all IBMQ systems that a user has access to.
    """
    ibmq_backends = {}
    for pro in Account.providers():
        for back in pro.backends():
            if not back.configuration().simulator:
                if ('alt' not in back.name()) \
                        and back.name().startswith('ibmq'):
                    if back.name() not in ibmq_backends:
                        ibmq_backends[back.name()] = [back]
                    else:
                        ibmq_backends[back.name()].append(back)
    return ibmq_backends



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
====================================================
Qiskit add-ons (:mod:`kaleidoscope.qiskit.overload`)
====================================================

.. currentmodule:: kaleidoscope.qiskit.overload

.. important::

   All of these routines are automatically integrated
   into Qiskit when using ``import kaleidoscope.qiskit``.


QuantumCircuit
==============

.. autosummary::
   :toctree: ../stubs/

   rshift
   transpile
   sample
   statevector
   unitary


Statevector
===========

.. autosummary::
   :toctree: ../stubs/

   bloch_components

"""

import kaleidoscope.qiskit.overload.circuit
import kaleidoscope.qiskit.overload.statevector

from kaleidoscope.qiskit.overload.circuit import (rshift, sample,
                                                  statevector, unitary,
                                                  transpile)

from kaleidoscope.qiskit.overload.statevector import bloch_components



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=wrong-import-position, too-many-return-statements

"""Functions that help for filtering reservations."""
import calendar
from datetime import datetime, timedelta, timezone
from kaleidoscope.errors import KaleidoscopeError


def time_intervals(interval):
    """Computes the start and ending datetimes for a set of commonly used intervals.

    All times are converted into the local timezone by applying the UTC offset
    for the locality.

    Current allowed ``interval`` values are:

    - 'now'
    - 'today'
    - 'tomorrow'
    - 'this week'
    - 'next week'
    - 'this month'
    - 'next month'

    Parameters:
        interval (str): A string representing the human readable interval.

    Returns:
        tuple: Start and end datetimes for specified time interval.

    Raises:
        KaleidoscopeError: Input string does not correspond to a known time interval.
    """
    now = datetime.now(timezone.utc)
    week_day = datetime.today().weekday()
    days_in_month = calendar.monthrange(now.year, now.month)[1]
    # Get UTC offset in hours
    dd = now.astimezone()
    utc_offset = dd.utcoffset() // timedelta(hours=1)

    today_start = datetime(year=now.year,
                           month=now.month,
                           day=now.day,
                           hour=0,
                           minute=0,
                           second=0)

    today_end = datetime(year=now.year,
                         month=now.month,
                         day=now.day,
                         hour=23,
                         minute=59,
                         second=59)

    if interval == 'now':
        return now, now
    elif interval == 'today':
        return now, today_end - timedelta(hours=utc_offset)
    elif interval == 'tomorrow':
        start = today_end + timedelta(seconds=1)
        end = datetime(year=start.year, month=start.month, day=start.day,
                       hour=23, minute=59, second=59)
        return start - timedelta(hours=utc_offset), end - timedelta(hours=utc_offset)
    elif interval == 'this week':
        start = now
        end = today_end + timedelta(days=6-week_day)
        return start, end - timedelta(hours=utc_offset)
    elif interval == 'next week':
        start = today_start - timedelta(days=week_day-7)
        end = today_end + timedelta(days=13-week_day)
        return start - timedelta(hours=utc_offset), end - timedelta(hours=utc_offset)
    elif interval == 'this month':
        start = now
        end = datetime(year=now.year, month=now.month, day=days_in_month,
                       hour=23, minute=59, second=59)
        return start, end - timedelta(hours=utc_offset)
    elif interval == 'next month':
        start = today_start + timedelta(days=days_in_month-now.day+1)
        _days_in_month = calendar.monthrange(now.year, start.month)[1]
        end = datetime(year=start.year, month=start.month, day=_days_in_month,
                       hour=23, minute=59, second=59)
        return start - timedelta(hours=utc_offset), end - timedelta(hours=utc_offset)
    else:
        raise KaleidoscopeError("Interval must be 'now', 'today', 'tomorrow', 'this week', \
                                'next week', 'this month', or 'next month'.")


def _get_reservations(systems, time_interval, collected_sys):
    """Get the reservations in the given time interval and
    add it to the collected reservations list.

    This is meant to be run in a different thread.

    Parameters:
        systems (list): A list of system instances.
        time_interval (tuple): A tuple of datetimes.
        collected_sys (list): A list of collected systems with reservations.
    """
    for system in systems:
        if any(system.reservations(*time_interval)):
            collected_sys.append(system)



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interactive Qsphere"""

import numpy as np
import scipy.linalg as la
import scipy.special as spsp
import matplotlib as mpl
import colorcet as cc
from plotly.subplots import make_subplots
import plotly.graph_objects as go
from kaleidoscope.errors import KaleidoscopeError
from .plotly_wrapper import PlotlyWidget, PlotlyFigure
from .bloch.primitives import BSPHERE


def qsphere(state, state_labels=True,
            state_labels_kind='bits',
            as_widget=False):
    """Plots a statevector of qubits using the qsphere
    representation.

    Parameters:
        state (ndarray): Statevector as 1D NumPy array.
        state_labels (bool): Show state labels.
        state_labels_kind (str): 'bits' (default) or 'ints'.
        as_widget (bool): Return a widget instance.

    Returns:
        PlotlyFigure or PlotlyWidget: Figure instance.

    Raises:
        KaleidoscopeError: Invalid statevector input.

    Example:

        .. jupyter-execute::

            from qiskit import QuantumCircuit
            from qiskit.quantum_info import Statevector
            import kaleidoscope.qiskit
            from kaleidoscope.interactive import qsphere

            qc = QuantumCircuit(3)
            qc.h(range(3))
            qc.ch(0,1)
            qc.s(2)
            qc.cz(2,1)
            state = qc.statevector()

            qsphere(state)
    """
    if state.__class__.__name__ in ['Statevector'] \
            and 'qiskit' in state.__class__.__module__:
        state = state.data

    if state.__class__.__name__ in ['DensityMatrix'] \
            and 'qiskit' in state.__class__.__module__:

        if not abs(1-state.data.dot(state.data).trace().real) < 1e-6:
            raise KaleidoscopeError('Input density matrix is not a pure state.')
        # pylint: disable=unexpected-keyword-arg
        state = la.eigh(state.data, subset_by_value=[0.99, 1.01])[1].ravel()

    if len(state.shape) == 2:
        if not abs(1-state.dot(state).trace().real) < 1e-6:
            raise KaleidoscopeError('Input density matrix is not a pure state.')
        # pylint: disable=unexpected-keyword-arg
        state = la.eigh(state, subset_by_value=[0.99, 1.01])[1].ravel()

    if len(state.shape) != 1:
        raise KaleidoscopeError('Input state is not 1D array.')

    if np.log2(state.shape[0]) % 1:
        raise KaleidoscopeError('Input is not a valid statevector of qubits.')

    eps = 1e-8
    norm = mpl.colors.Normalize(vmin=0, vmax=2*np.pi)
    cmap = cc.cm.CET_C1s
    num_qubits = int(np.log2(state.shape[0]))

    xvals = []
    yvals = []
    zvals = []
    colors = []
    bases = []
    probs = []
    marker_sizes = []

    for idx in range(2**num_qubits):
        prob = (state[idx]*state[idx].conj()).real
        if prob > eps:
            elem = bin(idx)[2:].zfill(num_qubits)
            weight = elem.count("1")
            zvalue = -2 * weight / num_qubits + 1
            number_of_divisions = spsp.comb(num_qubits, weight)
            weight_order = _bit_string_index(elem)
            angle = (float(weight) / num_qubits) * (np.pi * 2) + \
                    (weight_order * 2 * (np.pi / number_of_divisions))

            if (weight > num_qubits / 2) or (((weight == num_qubits / 2) and
                                              (weight_order >= number_of_divisions / 2))):
                angle = np.pi - angle - (2 * np.pi / number_of_divisions)

            xvalue = np.sqrt(1 - zvalue ** 2) * np.cos(angle)
            yvalue = np.sqrt(1 - zvalue ** 2) * np.sin(angle)

            bases.append(elem)
            probs.append(prob)
            xvals.append(xvalue)
            yvals.append(yvalue)
            zvals.append(zvalue)

            phase = np.arctan2(state[idx].imag, state[idx].real)
            phase = phase if phase >= 0 else phase+2*np.pi
            colors.append(mpl.colors.rgb2hex(cmap(norm(phase))))
            marker_sizes.append(np.sqrt(prob) * 40)

    if state_labels_kind == 'ints':
        bases = [int(kk, 2) for kk in bases]

    # Output figure instance
    fig = make_subplots(rows=5, cols=5,
                        specs=[[{"type": "scene", "rowspan": 5, "colspan": 5},
                                None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None,
                                {"rowspan": 1, "colspan": 1, "type": "domain"}
                                ]
                               ]
                        )

    figsize = (350, 350)

    # List for vector annotations, if any
    fig_annotations = []

    fig.add_trace(BSPHERE(), row=1, col=1)

    # latitudes
    for kk in _qsphere_latitudes(zvals):
        fig.add_trace(kk, row=1, col=1)

    fig.add_trace(go.Scatter3d(x=[0], y=[0], z=[0],
                               mode='markers',
                               opacity=0.6,
                               marker=dict(size=4,
                                           color='#555555'),
                               ),
                  row=1, col=1)

    for kk, _ in enumerate(xvals):
        fig.add_trace(go.Scatter3d(x=[0, xvals[kk]], y=[0, yvals[kk]], z=[0, zvals[kk]],
                                   mode="lines",
                                   hoverinfo=None,
                                   opacity=0.5,
                                   line=dict(color=colors[kk], width=3)
                                   ),
                      row=1, col=1
                      )

        if state_labels:
            xanc = 'center'
            if xvals[kk] != 0:
                if xvals[kk] < 0:
                    xanc = 'right'
                else:
                    pass

            yanc = 'middle'
            if zvals[kk] != 0:
                if zvals[kk] < 0:
                    yanc = 'top'
                else:
                    yanc = 'bottom'

            fig_annotations.append(dict(showarrow=False,
                                        x=xvals[kk]*1.1,
                                        y=yvals[kk]*1.1,
                                        z=zvals[kk]*1.1,
                                        text="<b>|{}\u3009</b>".format(bases[kk]),
                                        align='left',
                                        opacity=0.7,
                                        xanchor=xanc,
                                        yanchor=yanc,
                                        xshift=10,
                                        bgcolor="#ffffff",
                                        font=dict(size=10,
                                                  color="#000000",
                                                  ),
                                        )
                                   )

    fig.add_trace(go.Scatter3d(x=xvals, y=yvals, z=zvals,
                               mode='markers',
                               opacity=1,
                               marker=dict(size=marker_sizes,
                                           color=colors),
                               ),
                  row=1, col=1)

    slices = 128
    labels = ['']*slices
    values = [1]*slices

    phase_colors = [mpl.colors.rgb2hex(cmap(norm(2*np.pi*kk/slices))) for kk in range(slices)]

    fig.add_trace(go.Pie(labels=labels, values=values, hole=.6,
                         showlegend=False,
                         textinfo='none',
                         hoverinfo='none',
                         textposition="outside",
                         rotation=90,
                         textfont_size=12,
                         marker=dict(colors=phase_colors)
                         ),
                  row=5, col=5)

    pie_x = fig.data[-1]['domain']['x']
    pie_y = fig.data[-1]['domain']['y']

    fig['layout'].update(annotations=[
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='Phase',
            xanchor="center",
            yanchor="middle",
            showarrow=False,
            font=dict(size=9),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=pie_x[0]-0.03,
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='\U0001D70B',
            xanchor="left",
            yanchor="middle",
            showarrow=False,
            font=dict(size=14),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=pie_x[1]+0.03,
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='0',
            xanchor="right",
            yanchor="middle",
            showarrow=False,
            font=dict(size=12),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=pie_y[1]+0.05,
            text='\U0001D70B/2',
            xanchor="center",
            yanchor="top",
            showarrow=False,
            font=dict(size=12),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=pie_y[0]-0.05,
            text='3\U0001D70B/2',
            xanchor="center",
            yanchor="bottom",
            showarrow=False,
            font=dict(size=12),
        )
    ])

    fig.update_layout(width=figsize[0],
                      height=figsize[1],
                      autosize=False,
                      hoverdistance=50,
                      showlegend=False,
                      scene_aspectmode='cube',
                      margin=dict(r=15, b=15, l=15, t=15),
                      scene=dict(annotations=fig_annotations,
                                 xaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 yaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 zaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False)),
                      scene_camera=dict(eye=dict(x=0,
                                                 y=-1.4,
                                                 z=0.3)
                                        )
                      )

    if as_widget:
        return PlotlyWidget(fig)

    return PlotlyFigure(fig, modebar=True)


def _lex_index(n, k, lst):
    """Return  the lex index of a combination..

    Args:
        n (int): the total number of options .
        k (int): The number of elements.
        lst (list): list

    Returns:
        int: returns int index for lex order

    Raises:
        KaleidoscopeError: if length of list is not equal to k
    """
    if len(lst) != k:
        raise KaleidoscopeError("list should have length k")
    comb = list(map(lambda x: n - 1 - x, lst))
    dualm = sum([spsp.comb(comb[k - 1 - i], i + 1) for i in range(k)])
    return int(dualm)


def _bit_string_index(s):
    """Return the index of a string of 0s and 1s.

    Parameters:
        s (str): Bitstring.

    Returns:
        int: Index.

    Raises:
        KaleidoscopeError: If string is not binary.
    """
    n = len(s)
    k = s.count("1")
    if s.count("0") != n - k:
        raise KaleidoscopeError("s must be a string of 0 and 1")
    ones = [pos for pos, char in enumerate(s) if char == "1"]
    return _lex_index(n, k, ones)


def _qsphere_latitudes(zvals):
    """Latitude lines for sphere.

    Parameters:
        zvals (int): Input zvals

    Returns:
        list: List of Plotly traces.
    """
    lats = []
    u = np.linspace(0, 2*np.pi, 100)

    if 0 not in zvals:
        zvals = [0] + zvals
    for zv in zvals:
        th = np.arctan2(np.sqrt(1 - zv ** 2), zv)
        xvals = np.sin(th)*np.cos(u)
        yvals = np.sin(th)*np.sin(u)
        zvals = zv*np.ones_like(u)
        lats.append(go.Scatter3d(
            x=xvals, y=yvals, z=zvals,
            mode="lines",
            hoverinfo='skip',
            line=dict(
                color='#1e1e1e' if th == np.pi/2 else '#373737',
                width=2 if th == np.pi/2 else 1
            )
        ))
    return lats



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
# pylint: disable=unused-argument

"""Account functionality"""

from qiskit.exceptions import QiskitError
from qiskit.providers.ibmq.ibmqfactory import IBMQFactory
from ._config import set_default_provider, get_default_provider


def refresh(self):
    """Refresh the Account object
    """
    self.disable_account()
    self.load_account()

    # Trigger a refresh of the Systems provider
    from kaleidoscope.qiskit.services import Systems  # pylint: disable=cyclic-import
    Systems._refresh()


IBMQFactory.refresh = refresh
IBMQFactory.set_default_provider = set_default_provider
IBMQFactory.get_default_provider = get_default_provider

Account = IBMQFactory()
try:
    Account.load_account()
except QiskitError:
    pass



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=unexpected-keyword-arg

"""Module for creating device simulators automatically"""

import time
import warnings
import threading
import qiskit
from qiskit import Aer
from qiskit.providers import BaseBackend
from qiskit.providers.aer.noise import NoiseModel
from qiskit.providers.models import QasmBackendConfiguration, PulseBackendConfiguration
from ._account import Account


def _version2int(version_string):
    str_list = version_string.split(
        "-dev")[0].split("rc")[0].split("a")[0].split("b")[0].split(
            "post")[0].split('.')
    return sum([int(d if len(d) > 0 else 0) * (100 ** (3 - n))
                for n, d in enumerate(str_list[:3])])


AER_VERSION = _version2int(qiskit.__qiskit_version__['qiskit-aer'])


class _Credentials():
    def __init__(self, token='', url=''):
        self.token = token
        self.url = url
        self.hub = 'hub'
        self.group = 'group'
        self.project = 'project'


class DeviceSimulator(BaseBackend):
    """This is a simulator for a real IBMQ system.

    This class can be used as a drop in replacement for a IBMQ quantum
    device.
    """

    def __init__(self, backend, backend_name, local=True):
        """FakeBackend initializer.

        Args:
            backend (IBMQBackend): IBMQBackend instance
            backend_name (str): The name to give the backend.
            local (bool): Determines if Aer of IBMQ simulator should be used.
        """
        if backend.configuration().open_pulse:
            config = PulseBackendConfiguration.from_dict(backend.configuration().to_dict())
        else:
            config = QasmBackendConfiguration.from_dict(backend.configuration().to_dict())
        super().__init__(config)
        self._credentials = _Credentials()
        self._properties = backend.properties()
        self._configuration.simulator = True
        self._configuration.local = local
        self._configuration.backend_name = backend_name
        if AER_VERSION >= 60000:
            self.noise_model = NoiseModel.from_backend(self, warnings=False)
        else:
            self.noise_model = NoiseModel.from_backend(self)
        if local:
            self.sim = Aer.get_backend('qasm_simulator')
        else:
            pro = Account.get_provider(hub='ibm-q', group='open', project='main')
            self.sim = pro.backends.ibmq_qasm_simulator

    def properties(self):
        """Return backend properties"""
        return self._properties

    def run(self, qobj, **kwargs):  # pylint: disable=unused-argument
        job = self.sim.run(qobj, noise_model=self.noise_model)
        return job

    def jobs(self, **kwargs):  # pylint: disable=unused-argument
        """Fake a job history"""
        return []


def get_ibmq_systems():
    """Get instances for all IBMQ systems that the user has access to.

    Returns:
        dict: A dict of all IBMQ systems that a user has access to.
    """
    ibmq_backends = {}
    for pro in Account.providers():
        for back in pro.backends():
            if not back.configuration().simulator:
                if back.name() not in ibmq_backends \
                    and ('alt' not in back.name()) \
                        and back.name().startswith('ibmq'):
                    ibmq_backends[back.name()] = back
    return ibmq_backends


def _system_loader(service):
    if not any(Account.providers()):
        try:
            Account.load_account()
        except Exception:  # pylint: disable=broad-except
            pass
    systems = get_ibmq_systems()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for name, system in systems.items():
            if system.properties():
                new_name = '{}_'+name.split('_')[-1]+'_simulator'
                system = DeviceSimulator(system,
                                         new_name.format('aer'),
                                         local=True)
                system2 = DeviceSimulator(system,
                                          new_name.format('ibmq'),
                                          local=False)

                setattr(service, system.name(), system)
                setattr(service, system2.name(), system2)

    service.refreshing = False


class KaleidoscopeSimulatorService():
    """A service for IBMQ device simulators.

    Simulators are constructed from noise models that correspond
    to the latest calibration data returned from the devices.

    System simulators are loaded async, and the name corresponds to
    which underlying simulator is doing the computation, e.g. :code:`aer_*`
    or :code:`ibmq_*`.

    Systems are attached to the service as attributes and the service
    object is available at the top-level via :code:`simulators`.
    For example, a :code:`ibmq_vigo` simulator using :code:`Aer` can be retrieved
    via:

    .. jupyter-execute::

        from kaleidoscope.qiskit import Simulators
        sim = Simulators.aer_vigo_simulator

    Attributes:
        refreshing (bool): Is the service refreshing its simulators async.

    """
    def __init__(self):
        self.refreshing = False
        self.refresh()

    def __call__(self):
        return list(vars(self).keys())

    def __getattr__(self, attr):
        if 'aer' in attr or 'ibmq' in attr:
            while self.refreshing:
                time.sleep(0.1)
                if attr in self.__dict__:
                    return self.__dict__[attr]
            if attr not in self.__dict__:
                raise AttributeError("Couldn't load {}.".format(attr))
            return self.__dict__[attr]
        else:
            raise AttributeError("Simulators does not have attribute {}".format(attr))

    def refresh(self):
        """Refresh the service for new backends if IBMQ
        account was not loaded before init.
        """
        self.refreshing = True
        thread = threading.Thread(target=_system_loader,
                                  args=(self, ))
        thread.start()



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.

"""Configrc file functionality"""

import os
import warnings
import configparser
from kaleidoscope.errors import KaleidoscopeError


def has_kal_rc():
    """
    Checks to see if the kalrc file exists in the default
    location, i.e. HOME/.kaleidoscope/kalrc
    """
    kal_conf_dir = os.path.join(os.path.expanduser("~"), '.kaleidoscope')
    if os.path.exists(kal_conf_dir):
        kal_rc_file = os.path.join(kal_conf_dir, 'kalrc')
        rc_exists = os.path.isfile(kal_rc_file)
        if rc_exists:
            return True, kal_rc_file
        else:
            return False, None
    else:
        return False, None


def generate_kal_rc():
    """
    Generate a blank kalrc file.
    """
    # Check for write access to home dir
    if not os.access(os.path.expanduser("~"), os.W_OK):
        return False
    kal_conf_dir = os.path.join(os.path.expanduser("~"), '.kaleidoscope')
    if not os.path.exists(kal_conf_dir):
        try:
            os.mkdir(kal_conf_dir)
        except Exception:  # pylint: disable=broad-except
            warnings.warn('Cannot write config file to user home dir.')
            return False
    kal_rc_file = os.path.join(kal_conf_dir, 'kalrc')
    rc_exists = os.path.isfile(kal_rc_file)
    if rc_exists:
        # Do not overwrite
        return False
    else:
        # Write a basic file with qutip section
        cfgfile = open(kal_rc_file, 'w')
        config = configparser.ConfigParser()
        config.add_section('kaleidoscope')
        config.write(cfgfile)
        cfgfile.close()
        return True


def has_rc_key(rc_file, key):
    """Checks if a given key exists in
    a given rc_file.

    Parameters:
        rc_file (str): Specified rc file.
        key (str): The key.

    Returns:
        bool: Does the key exist.

    Raises:
        NoSectionError: Kaleidoscope section does
                        not exist.
    """
    config = configparser.ConfigParser()
    config.read(rc_file)
    if config.has_section('kaleidoscope'):
        opts = config.options('kaleidoscope')
        return key in opts
    else:
        raise configparser.NoSectionError('kaleidoscope')


def write_rc_key(rc_file, key, value):
    """
    Writes a single key value to the qutiprc file

    Parameters:
        rc_file (str): String specifying file location.
        key (str): The key name to be written.
        value (str): Value corresponding to given key.
    """
    if not os.access(os.path.expanduser("~"), os.W_OK):
        return
    cfgfile = open(rc_file, 'w')
    config = configparser.ConfigParser()
    if not config.has_section('kaleidoscope'):
        config.add_section('kaleidoscope')
    config.set('kaleidoscope', key, value)
    config.write(cfgfile)
    cfgfile.close()


def get_rc_key(rc_file, key):
    """Get the desired key from an kal rc_file

    Parameters:
        rc_file (str): String specifying file location.
        key (str): The key name to be written.

    Returns:
        str: The key value.

    Raises:
        NoSectionError: rc_file missing kaleidoscope section.
        KaleidoscopeError: Key not found.
    """
    config = configparser.ConfigParser()
    config.read(rc_file)
    if not config.has_section('kaleidoscope'):
        raise configparser.NoSectionError('kaleidoscope')

    opts = config.options('kaleidoscope')
    if key not in opts:
        raise KaleidoscopeError('Key not found.')
    return config.get('kaleidoscope', key)



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Overloading of the Statevector class in Qiskit"""

from qiskit.quantum_info.states.statevector import Statevector
from kaleidoscope.interactive.bloch.utils import bloch_components as bcomp


def bloch_components(self):
    """Returns the Bloch components of the statevector.

    Example:

        .. jupyter-execute::

            from qiskit import QuantumCircuit
            import kaleidoscope.qiskit

            qc = QuantumCircuit(3)
            qc.h(range(3))
            qc.ch(0,1)
            qc.s(2)
            qc.cz(2,1)

            qc.statevector().bloch_components()
    """
    return bcomp(self.data)


# Add methods to Statevector class
Statevector.bloch_components = bloch_components



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
# pylint: disable=unused-argument

"""Configrc file functionality"""

from qiskit.providers.ibmq.accountprovider import AccountProvider
from kaleidoscope.configrc import has_kal_rc, has_rc_key, write_rc_key, get_rc_key
from kaleidoscope.errors import KaleidoscopeError


def set_default_provider(self, hub=None, group=None, project=None, overwrite=False):
    """Set the default provider for IBM Q systems.

    Parameters:
        hub (str, AccountProvider): A hub name, or Qiskit provider instance
        group (str): A group name.
        project (str): A project name.
        overwrite (bool): Overwrite if already set.

    Raises:
        KaleidoscopeError: Input not a valid provider.
        KaleidoscopeError: Could not load kalrc.
        KaleidoscopeError: Default provider found and overwrite=False.
    """
    if isinstance(hub, AccountProvider):
        _temp = hub
        hub = _temp.credentials.hub
        group = _temp.credentials.group
        project = _temp.credentials.project
    else:
        pro = self.providers(hub, group, project)
        if not pro:
            raise KaleidoscopeError('Input hub, group, and/or project not valid.')
        if len(pro) > 1:
            raise KaleidoscopeError('Inputs do not specify a unique provider.')
        pro = pro[0]
        hub = pro.credentials.hub
        group = pro.credentials.group
        project = pro.credentials.project

    provider_str = "//".join([hub, group, project])

    has_rc, rc_file = has_kal_rc()
    if not has_rc:
        raise KaleidoscopeError('Could not load kalrc.')

    if has_rc_key(rc_file, 'default_provider'):
        if not overwrite:
            raise KaleidoscopeError('Default provider already set, use overwrite=True.')

    write_rc_key(rc_file, 'default_provider', provider_str)

    # Trigger a refresh of the Systems provider
    from kaleidoscope.qiskit.services import Systems  # pylint: disable=cyclic-import
    Systems._refresh()


def get_default_provider(self):
    """Return the default provider, if any.

    Returns:
        str: Default provider.

    Raises:
        KaleidoscopeError: Could not load rcfile.
    """
    has_rc, rc_file = has_kal_rc()
    if not has_rc:
        raise KaleidoscopeError('Could not load kalrc.')
    return get_rc_key(rc_file, 'default_provider')



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=unused-import

"""Tests for system filtering"""

from kaleidoscope.qiskit import Systems


def test_num_qubits_filter():
    """Tests filtering by number of qubits"""
    systems = Systems.num_qubits == 5
    for system in systems:
        assert system.configuration().num_qubits == 5


def test_num_qubits_filter2():
    """Tests filtering by number of qubits 2"""
    systems = Systems.num_qubits > 5
    for system in systems:
        assert system.configuration().num_qubits > 5


def test_openpulse_filter():
    """Tests filtering by pulse access"""
    systems = Systems.open_pulse
    for system in systems:
        assert system.configuration().open_pulse


def test_qv_filter():
    """Tests filtering by quantum volume"""
    systems = Systems.quantum_volume >= 16
    for system in systems:
        assert system.configuration().quantum_volume >= 16



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
# pylint: disable=unused-argument

"""Backend filtering functionality"""

import threading
from datetime import datetime, timedelta
from qiskit.providers.ibmq.ibmqbackend import IBMQBackend
from kaleidoscope.errors import KaleidoscopeError
from kaleidoscope.utils.hardware import local_hardware_info
from .reservation_helpers import (time_intervals, _get_reservations)

_NCPUS = local_hardware_info()['cpus']


class BackendCollection(list):
    """A list subclass that makes handling IBMQBackends easier.
    """
    def __init__(self, data):
        """BackendCollection constructor

        Parameters:
            data (list or BackendCollection): List of IBMQbackend instances.

        Raises:
            TypeError: Must be list of IBMQBackends only.
        """
        for dd in data:
            if not isinstance(dd, IBMQBackend):
                raise TypeError('Backend collection requires IBMQBackend instances.')
        super().__init__(data)

    def __and__(self, other):
        return BackendCollection(set(self) & set(other))

    def __or__(self, other):
        return BackendCollection(set(self) | set(other))

    def __add__(self, other):
        """Adds to BackendCollections together.
        """
        if not isinstance(other, (BackendCollection, list)):
            if isinstance(other, list):
                other = BackendCollection(other)
            else:
                raise TypeError('BackendCollection addition works only for collections and lists.')

        return BackendCollection(set(self) | set(other))

    def __radd__(self, other):
        """Adds to BackendCollections together.
        """
        if not isinstance(other, (BackendCollection, list)):
            if isinstance(other, list):
                other = BackendCollection(other)
            else:
                raise TypeError('BackendCollection addition works only for collections and lists.')

        return BackendCollection(set(self) | set(other))

    def __getattr__(self, name):
        if name == 'num_qubits':
            out = NumQubits(self)
        elif name == 'open_pulse':
            out = HasPulse(self)
        elif name == 'quantum_volume':
            out = QVCompare(self)
        elif name == 'operational':
            out = IsOperational(self)
        elif name == 'max_circuits':
            out = MaxCircuits(self)
        elif name == 'max_shots':
            out = MaxShots(self)
        elif name == 'reservations':
            out = Reservations(self)
        else:
            raise AttributeError("BackendCollection does not have attr '{}'.".format(name))
        return out


class Comparator(BackendCollection):
    """A skeleton constructor class built on BackendCollection.
    """
    def __eq__(self, other):
        raise KaleidoscopeError('Not implimented')

    def __ne__(self, other):
        raise KaleidoscopeError('Not implimented')

    def __gt__(self, other):
        raise KaleidoscopeError('Not implimented')

    def __ge__(self, other):
        raise KaleidoscopeError('Not implimented')

    def __lt__(self, other):
        raise KaleidoscopeError('Not implimented')

    def __le__(self, other):
        raise KaleidoscopeError('Not implimented')


class IsOperational(Comparator):
    """Implements a is operational check.
    """
    def __init__(self, data):
        """BackendCollection constructor

        Parameters:
            data (BackendCollection): Collection of IBMQbackend instances
        """
        self._full_data = data
        super().__init__([back for back in data if back.status().operational])

    def __eq__(self, other):
        if (other is not None) and not isinstance(other, (bool, int)):
            raise KaleidoscopeError('Can only compare against boolean and int values.')
        if isinstance(other, int) and other not in [0, 1]:
            raise KaleidoscopeError('Integer comparison must be against 0 or 1.')
        out = [back for back in self._full_data if back.status().operational == other]
        return BackendCollection(out)

    def __ne__(self, other):
        if (other is not None) and not isinstance(other, (bool, int)):
            raise KaleidoscopeError('Can only compare against boolean and int values.')
        if isinstance(other, int) and other not in [0, 1]:
            raise KaleidoscopeError('Integer comparison must be against 0 or 1.')
        out = [back for back in self._full_data if back.status().operational != other]
        return BackendCollection(out)


class HasPulse(Comparator):
    """Implements a open_pulse check.
    """
    def __init__(self, data):
        """BackendCollection constructor

        Parameters:
            data (BackendCollection): Collection of IBMQbackend instances
        """
        self._full_data = data
        super().__init__([back for back in data if back.configuration().open_pulse])

    def __eq__(self, other):
        if (other is not None) and not isinstance(other, (bool, int)):
            raise KaleidoscopeError('Can only compare against boolean and int values.')
        if isinstance(other, int) and other not in [0, 1]:
            raise KaleidoscopeError('Integer comparison must be against 0 or 1.')
        out = [back for back in self._full_data if back.configuration().open_pulse == other]
        return BackendCollection(out)

    def __ne__(self, other):
        if (other is not None) and not isinstance(other, (bool, int)):
            raise KaleidoscopeError('Can only compare against boolean and int values.')
        if isinstance(other, int) and other not in [0, 1]:
            raise KaleidoscopeError('Integer comparison must be against 0 or 1.')
        out = [back for back in self._full_data if back.configuration().open_pulse != other]
        return BackendCollection(out)


class NumQubits(Comparator):
    """Implements a number of qubits comparator.
    """
    def __eq__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits == other]
        return BackendCollection(out)

    def __ne__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits != other]
        return BackendCollection(out)

    def __gt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits > other]
        return BackendCollection(out)

    def __ge__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits >= other]
        return BackendCollection(out)

    def __lt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits < other]
        return BackendCollection(out)

    def __le__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().num_qubits <= other]
        return BackendCollection(out)


class MaxCircuits(Comparator):
    """Implements a number of max circuits comparator.
    """
    def __eq__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments == other]
        return BackendCollection(out)

    def __ne__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments != other]
        return BackendCollection(out)

    def __gt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments > other]
        return BackendCollection(out)

    def __ge__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments >= other]
        return BackendCollection(out)

    def __lt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments < other]
        return BackendCollection(out)

    def __le__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_experiments <= other]
        return BackendCollection(out)


class MaxShots(Comparator):
    """Implements a number of max shots comparator.
    """
    def __eq__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots == other]
        return BackendCollection(out)

    def __ne__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots != other]
        return BackendCollection(out)

    def __gt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots > other]
        return BackendCollection(out)

    def __ge__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots >= other]
        return BackendCollection(out)

    def __lt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots < other]
        return BackendCollection(out)

    def __le__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = [back for back in self if back.configuration().max_shots <= other]
        return BackendCollection(out)


class QVCompare(Comparator):
    """Implements a quantum volume comparator.
    """
    def __eq__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val == other:
                    out.append(back)
        return BackendCollection(out)

    def __ne__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val != other:
                    out.append(back)
        return BackendCollection(out)

    def __gt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val > other:
                    out.append(back)
        return BackendCollection(out)

    def __ge__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val >= other:
                    out.append(back)
        return BackendCollection(out)

    def __lt__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val < other:
                    out.append(back)
        return BackendCollection(out)

    def __le__(self, other):
        if not isinstance(other, int):
            raise KaleidoscopeError('Can only compare against ints')
        out = []
        for back in self:
            if hasattr(back.configuration(), 'quantum_volume'):
                qv_val = back.configuration().quantum_volume
                if qv_val and qv_val <= other:
                    out.append(back)
        return BackendCollection(out)


class Reservations(Comparator):
    """Implements a upcoming reservations comparator.
    """

    def get_reservations(self, time_interval):
        """Gets the reservations in the given time interval.

        Parameters:
            time_interval (tuple): Datetime objects for interval.

        Returns:
            BackendCollection: Collection of backends with reservations.
        """
        num_sys = len(self)
        grouped_systems = [self[i:i+_NCPUS] for i in range(0, num_sys, _NCPUS)]

        threads = []
        collected_sys = []
        for kk, _ in enumerate(grouped_systems):
            thread = threading.Thread(target=_get_reservations,
                                      args=(grouped_systems[kk], time_interval, collected_sys))
            thread.start()
            threads.append(thread)

        for kk in range(len(grouped_systems)):
            threads[kk].join()

        return BackendCollection(collected_sys)

    def __eq__(self, other):
        if not isinstance(other, (str, datetime)):
            raise KaleidoscopeError('Equality can only be checked against a string or datetime')

        if isinstance(other, datetime):
            tvals = (other, other)
        else:
            tvals = time_intervals(other)

        return self.get_reservations(tvals)

    def __lt__(self, other):
        if not isinstance(other, (str, datetime)):
            raise KaleidoscopeError('Equality can only be checked against a string or datetime')

        if isinstance(other, datetime):
            if other <= datetime.now():
                raise KaleidoscopeError('Datetime must be in the future.')
            tvals = (None, other-timedelta(seconds=1))
        else:
            tvals = (None, time_intervals(other)[0]-timedelta(seconds=1))

        return self.get_reservations(tvals)

    def __le__(self, other):
        if not isinstance(other, (str, datetime)):
            raise KaleidoscopeError('Equality can only be checked against a string or datetime')

        if isinstance(other, datetime):
            tvals = (None, other)
        else:
            tvals = (None, time_intervals(other)[1])

        return self.get_reservations(tvals)

    def __gt__(self, other):
        if not isinstance(other, (str, datetime)):
            raise KaleidoscopeError('Equality can only be checked against a string or datetime')

        if isinstance(other, datetime):
            if other <= datetime.now():
                raise KaleidoscopeError('Datetime must be in the future.')
            tvals = (other+timedelta(seconds=1), None)
        else:
            tvals = (time_intervals(other)[1]+timedelta(seconds=1), None)

        return self.get_reservations(tvals)

    def __ge__(self, other):
        if not isinstance(other, (str, datetime)):
            raise KaleidoscopeError('Equality can only be checked against a string or datetime')

        if isinstance(other, datetime):
            if other <= datetime.now():
                raise KaleidoscopeError('Datetime must be in the future.')
            tvals = (other, None)
        else:
            tvals = (time_intervals(other)[0], None)

        return self.get_reservations(tvals)



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=wrong-import-position

"""
==================================================================
IBM Quantum Services (:mod:`kaleidoscope.qiskit.services`)
==================================================================

.. currentmodule:: kaleidoscope.qiskit.services

.. autosummary::
   :toctree: ../stubs/

   Account
   KaleidoscopeSystemService
   KaleidoscopeSimulatorService

"""
from ._account import Account
from ._systems import KaleidoscopeSystemService
from ._simulators import KaleidoscopeSimulatorService

Systems = KaleidoscopeSystemService()
Simulators = KaleidoscopeSimulatorService()
